unit ft_scanthread;

// FTP Client Thread Code
// (c)2000 Chris Lautenbach (viper@egate.net)
// Based on Synapse Synchronous TCP/IP Library for Delphi by Lukas Gebauer (gebauerl@mlp.cz)
// Written in Delphi 5 (tested on no other versions)

{

This unit contains ft_scanthread, which is a self-contained FTP client in a thread.  This permits
downloading of multiple files from multiple FTP sites simultaneously.  It works relatively well, although
it only parses UNIX type directory listings (who uses anything else? :).  It uses PASV mode exclusively
for file and directory list transfers so it should work fine through even the most primitive NAT
implementations.  There is extensive error handling including timeouts, automatic re-establishment of
data port connections during PASV connection operations, checks for bag logins, etc.

The source isn't super-documented, but should be relatively straightforward for anyone with an understanding
of the FTP protocol and blocking sockets.  (And it's free ;)

Notes:

1. This code uses TurboPower SysTools StStrL unit for string processing.  I've included a DEFINE if you
   don't have this amazing library.  Enable the NOSYSTOOLS define (default) if you don't have it, and
   it'll use internal string processing routines (which probably aren't as fast as the SysTools routines,
   but they do work).

2. Basically, the thread is capable of the following functions:

   1. Changing directories.
   2. Listing directory contents.
   3. Downloading files (including resumed transfers via REST command).
   4. Checking if a site is up/can be logged into/etc.
   5. Complete logging of the entire FTP session to a text file (make a subdirectory called logs/ in
      the directory your EXE is located).

   It's probably only useful for fetching files that you already know the paths for, as there is
   no automatic mechanism for returning directory listings (eg. an interactive session would require
   some modification).

   Multiple threads do work perfectly, and the thread uses minimal CPU time while transferring.

3. In order to get a thread going, do the following:

   Create the thread object.
   Set the f_Host and f_Port values with the host and port of the ftp server.
   Set the f_SID value with a unique identifier for the logfile.
   Set f_OwnerHWND with the handle of the window that owns the thread.
   Set Check_Status to TRUE or FALSE depending on the operation.  If TRUE, the thread will just connect
     to the specified site and log out to make sure it's up.
   Set the OnTerminate event handler.

   If you wish to download a single file, put its full name/path in the f_retr variable, and the local
     path/filename in f_lretr.
   If you wish to download multiple files, set f_retr to 'multi' and assign a list of fully validated
     pathnames in f_flist (TStringList).  The f_flist entries should be formatted like so:

     blah*/pub/incoming/test.zip*c:\my documents\test.zip

     Each line contains three elements delimited by asterisks.  The first is an identifier (not used by
     this code, use whatever you want, just don't leave it blank :) the second is the remote filename/path
     and the the third is the local filename/path.

   During file transfer operations, a file is automatically resumed if the local file exists and is
   less than the size of the remote file.

4. Status updates are sent to the calling window via windows messages, WM_MSGUPDATE, WM_STATUSUPDATE, and
   WM_SPEEDUPDATE.

   WM_MSGUPDATE is just a method for passing debug information (strings).
   WM_STATUSUPDATE sends information about the current operation (transfer/change dir/etc) plus current
     transfer rate (in string form).  This information is suitable for being displayed in a TListView when
     you have multiple transfer threads running and want to keep track of them all.
   WM_SPEEDUPDATE just sends the current transfer rate as a real.  This was put in place so you could
     easily display the total throughput of all the active threads without having to parse the
     WM_STATUSUPDATE strings to get the information.

   If the f_OwnerHWND variable has not been set, no status messages are sent.

Enjoy!

}

{$DEFINE NOSYSTOOLS}

interface

uses
  Windows, SysUtils, Classes, comctrls, BlckSock, Synautil, Messages{$IFNDEF NOSYSTOOLS}, StStrL{$ENDIF};

const WM_MSGUPDATE = WM_USER + 201;
      WM_STATUSUPDATE = WM_USER + 202;
      WM_SPEEDUPDATE = WM_USER + 203;

type charset = set of char;

     update_rec = record
       si_0, si_1, si_2, si_3 : string;
       tl : tlistitem;
     end;

     speed_update_rec = record
       threadid : integer;
       speed : real;
     end;

  FTPScanThread = class(TThread)
  private
    { Private declarations }
    ControlSocket, DataSocket : TTCPBlockSocket;
  protected
    procedure Execute; override;

    // Parsing code

{$IFDEF NOSYSTOOLS}
    function  ChInStr(ch : char; st : string) : boolean;
    function  extractwordl(n : byte; s : string; delims : string) : string;
    function  wordcountl(s : string; delims : string) : byte;
{$ENDIF}
    function  IsNumeric(st : string) : boolean;
    function  StripChars(St : string; StripList : CharSet) : string;
    function  ReplaceChar(St : string; ChIn, Chout : char):string;
    function  ParseFileName(St : string) : string;
    function  ParseUserName(St : string) : string;
    function  ParseFileSize(St : string) : integer;

    // Low level FTP code

    procedure SendControl(St : string);
    function  WaitForCode(Code : string; TimeOut : Cardinal) : integer;
    procedure DoCommand(St : string);
    function  CheckSocketError : boolean;

    // High level FTP code (commands, etc)

    function  Login : boolean;
    function  SetPASV(RetryCnt : integer) : boolean;
    procedure RetrieveFile(FN,LFN : string);
    procedure ScanDirectory(Path : string);
    function  ChangeDirectory(Path : string) : boolean;
    function  GetDirectory(Strs : TStringList) : boolean;

    // Timer code

    procedure ResetTimeout(Secs : cardinal);
    function  CheckTimeout : boolean;

    // File code

    function  GetFileSize(FN : string) : integer;
    procedure Logentry(St : string);
    procedure DeleteFiles(FileSpec : string);

    // Interactivity code
    procedure updatemain;
    procedure SendMsg(Msg : string);
    procedure UpdateStatus;
    procedure UpdateSpeed(Speed:real);

  public
    f_SID, f_Host, f_User, f_Pass, f_Desc, f_ScanStatus,
      FileStatus, TimeOutStr, CurPath, CurState, UpdateMsg,
      ControlPartialBuffer, LastResponse, FoundURL : string;
    f_Port : word;
    f_OwnerHWND : HWND;
    f_ur : update_rec;
    f_su : speed_update_rec;
    f_tagptr : pointer;
    f_retr, f_lretr : string;
    f_flist : tstringlist;
    Check_Status, Disconnected, TimedOut : boolean;
    MainListItem : TListItem;
    ControlBuffer, MemBuffer : TStringList;
    MBFile, Logfile : textfile;
    DataFile : file;
    recursecnt, File_Size, BytesReceived, ErrCode,
      total_kbytes, total_files : integer;
    to_start, to_end : cardinal;

    constructor Create(CreateSuspended : boolean);
    destructor Destroy; override;
  end;

implementation

{ FTPScanThread }

constructor FTPScanThread.Create(CreateSuspended : boolean);
begin
  inherited Create(CreateSuspended);

  // Create sockets

  ControlSocket := TTCPBlockSocket.Create;
  ControlSocket.CreateSocket;
  ControlBuffer := TStringList.Create;

  DataSocket := TTCPBlockSocket.Create;
  DataSocket.CreateSocket;

  // Set up variables/init objects

  ErrCode := 0;
  FileStatus := '';
  f_OwnerHWND := 0;
  TimedOut := FALSE;
  MemBuffer := TStringList.Create;
end;

destructor FTPScanThread.Destroy;
begin
  // Free up allocated resources;

  if Assigned(ControlSocket) then
    ControlSocket.Free;

  if Assigned(ControlSocket) then
    DataSocket.Free;

  if Assigned(ControlSocket) then
    MemBuffer.Free;

  if Assigned(ControlSocket) then
    ControlBuffer.Free;

  inherited Destroy;
end;

{$IFDEF NOSYSTOOLS}
function FTPScanThread.ChInStr(ch : char; st : string) : boolean;
// returns TRUE if the character ch is found in the string st
var idx : integer;
begin
 Result := FALSE;
  for idx := 1 to length(st) do
    if st[idx] = ch then
    begin
      Result := TRUE;
      Exit;
    end;
end;

function FTPScanThread.wordcountl(s : string; delims : string) : byte;
// show word count
var count, idx, len : integer;
begin
  count := 0;
  idx := 1;
  len := Length(S);
  while idx <= len do
  begin
    while (idx <= len) and chinstr(s[idx],delims) do inc(idx);
    if idx <= len then inc(count);
    while (idx <= len) and not(chinstr(s[idx],delims)) do inc(idx);
  end;
  Result := count;
end;

function FTPScanThread.extractwordl(n : byte; s : string; delims : string) : string;
// extract word n from string s, using delims as word delimiters
var count, idx, len : integer;
    TempSt : string;
begin
  count := 0;
  idx := 1;
  TempSt := '';
  len := Length(S);
  while (idx <= len) and (count <> N) do
  begin
    while (idx <= len) and (chinstr(s[idx],delims)) do inc(idx);
    if idx <= len then inc(count);
    while (idx <= len) and (not chinstr(s[idx],delims)) do
    begin
      if count = N then TempSt := TempSt + S[idx];
      Inc(idx);
    end;
  end;
  Result := TempSt;
end;
{$ENDIF}

procedure FTPScanThread.ResetTimeout(Secs : cardinal);
begin
  // Resets the timeout timer to expire in Secs seconds.
  TimedOut := FALSE;
  to_end := GetTickCount + (Secs * 1000);
  to_start := GetTickCount;
end;

function FTPScanThread.CheckTimeout : boolean;
var nt : string;
begin
  // Checks to see if the timer has expired, returns TRUE if timed out
  TimedOut := (GetTickCount > to_end);
  CheckTimeout := TimedOut;
  nt := inttostr((GetTickCount - to_start) DIV 1000);
end;

function FTPScanThread.GetFileSize(FN : string) : integer;
var TSR : TSearchRec;
begin
  // Get the size of the file FN
  If FindFirst(FN,faAnyFile,TSR) = 0 then
    Result := TSR.Size
      ELSE
    Result := -1;
  FindClose(TSR);
end;

function FTPScanThread.StripChars(St : string; StripList : CharSet) : string;
  { Strip chars in StripList from string St }
var IDX : integer;
    OutSt : string;
begin
  OutSt := '';
  For IDX := 1 to Length(St) do
    If NOT (St[IDX] in StripList) then
      OutSt := OutSt + St[IDX];
  StripChars := OutSt;
end;

function FTPScanThread.ReplaceChar(St : string; ChIn, Chout : char):string;
var idx : byte;
    outst : string;
begin
  outst:=st;
  for idx := 1 to length(St) do if outst[idx]=chin then outst[idx]:=chout;
  ReplaceChar:=OutSt;
end;

function FTPScanThread.IsNumeric(st : string) : boolean;
var idx : integer;
begin
  Result := TRUE;
  for idx := 1 to length(st) do
    if NOT (st[idx] in ['0'..'9']) then
    begin
      Result := FALSE;
      Exit;
    end;
end;

function FTPScanThread.ParseFileName(St : string) : string;
var IDX, Cnt : integer;
    OutSt : string;
begin
  If WordCountL(St,#32) < 8 then
  begin
    Result := '';
    Exit;
  end;
  Cnt := 0;
  OutSt := '';
  For IDX := 1 to Length(St) do
  begin
    If (St[IDX] <> #32) AND (St[IDX-1] = #32) then Inc(Cnt);
    If Cnt >= 8 then
      OutSt := OutSt + St[IDX];
  end;
  Result := Trim(OutSt);
end;

function FTPScanThread.ParseUserName(St : string) : string;
begin
  ParseUserName := Trim(ExtractWordl(3,St,#32));
end;

function FTPScanThread.ParseFileSize(St : string) : integer;
var tempst : string;
begin
  If WordCountL(St,#32) < 8 then
    Result := -1
      ELSE
    begin
      tempst := Trim(ExtractWordL(5,St,#32));
      If IsNumeric(tempst) then
        Result := StrToInt(tempst)
          ELSE
        Result := -1;
    end;
end;

procedure FTPScanThread.SendControl(St : string);
var OutBuf : array[1..1024] of char;
    IDX, OutLen : integer;
begin
  OutLen := Length(St);

  For IDX := 1 to OutLen do
    OutBuf[IDX] := St[IDX];

  ControlSocket.SendBuffer(@OutBuf,OutLen);
end;

procedure FTPScanThread.LogEntry(St : String);
begin
  writeln(logfile,st);
  flush(logfile);
end;

procedure FTPScanThread.DoCommand(St : string);
begin
  If (ErrCode <> 0) OR Terminated OR TimedOut then Exit;
  SendControl(St + #13+#10);
  logentry('# '+St);
end;

function FTPScanThread.WaitForCode(Code : string; TimeOut : Cardinal) : integer;
var TempSt : string;
    idx,{ int1, int2,} NumCodes, CodeFound : integer;
begin
  Result := 0;
  If TimedOut OR Terminated then Exit;

  CodeFound := 0;

  NumCodes := WordCountL(Code, '|');

  ResetTimeout(Timeout);

  repeat
    If (ControlSocket.LastError = 0) then TempSt := ControlSocket.Recvstring(30);

    If TempSt <> '' then ControlBuffer.Add(TempSt);

    If (ControlBuffer.Count > 0) then // pending messages, process them
    begin
      // Data has been received, reset timer

      ResetTimeout(Timeout);

      // Process incoming line of text

      TempSt := ControlBuffer[0];
      ControlBuffer.Delete(0);

      For idx := 1 to NumCodes do
        If (Copy(TempSt, 1, Length(ExtractWordl(idx,Code,'|'))) = ExtractWordl(idx,Code,'|')) then
        begin
          CodeFound := IDX;
          Break;
        end;

      logentry('% '+TempSt);

      LastResponse := TempSt;
    end;
    CheckTimeout;
    Sleep(25);
  until (ControlSocket.LastError <> 0) OR (CodeFound > 0) OR (ErrCode <> 0) OR Terminated OR TimedOut;
  IF (ControlSocket.LastError <> 0) then Logentry('! Control socket error:'+geterrordesc(controlsocket.lasterror));
  If TimedOut then
  begin
    Logentry('! Timed out waiting for code(s): ' + Code);
    Logentry('! Buffer contents: "'+ControlPartialBuffer+'"');
  end;

  WaitForCode := CodeFound;
end;

function  FTPScanThread.SetPASV(RetryCnt : integer) : boolean;
var curcnt, int1, int2, int3 : integer;
    tst1,tst2:string;
begin
  // Set up PASV connection
  CurCnt := 0;
  repeat
    inc(CurCnt);
    LogEntry('! PASV: Connect attempt #'+inttostr(CurCnt));
    DoCommand('PASV');
    Result := (WaitForCode('227 ',30) = 1);
    If Result then
    begin
      bytesreceived := 0;
      tst1 := extractwordl(5, LastResponse, ',)');
      tst2 := extractwordl(6, LastResponse, ',)');
      if isnumeric(tst1) and isnumeric(tst2) then
      begin
        int1 := strtoint(tst1);
        int2 := strtoint(tst2);
        int3 := (int1 * 256) + int2;
        LogEntry('! PASV: Connect to port '+inttostr(int3));
        DataSocket.CloseSocket;
        DataSocket.CreateSocket;
        DataSocket.Connect(f_host,inttostr(int3));
        Result := (DataSocket.LastError = 0);
        If NOT Result then
          LogEntry('! PASV: Connection Error (' +geterrordesc(datasocket.lasterror)+')');
      end
        ELSE
      begin
        logentry('! PASV: Incorrect response received ('+lastresponse+')');
      end;
    end
      ELSE
    begin
      LogEntry('! PASV: Failed to create connection');
      TimedOut := TRUE;
    end;
  until Result OR TimedOut OR (CurCnt > RetryCnt);
  If (CurCnt > RetryCnt) then
  begin
    LogEntry('! PASV: Reconnect attempt limit exceeded, aborting thread');
    Result := FALSE;
    Terminate;
  end;
end;

procedure FTPScanThread.RetrieveFile(FN,LFN : string);
var Response : integer;
    databuf : array[1..65336] of char;
    LocalSize, AmtRead : integer;
    StartCnt, TimeCnt : cardinal;
    kbs, timetaken : currency;
begin
  If TimedOut OR Terminated then Exit;

//  sendmsg('retrieving: '+fn);

  DoCommand('SIZE ' + FN);
  Response := WaitForCode('213 |550 ',30);
  If Response = 2 then
  begin
    logentry('> access denied:'+fn);
    exit;
  end;

  File_Size := StrToInt(Trim(ExtractWordl(2,LastResponse,#32)));
  If File_Size = 0 then // Don't download 0 byte files
  begin
    logentry('! File size is 0 bytes, skipping.');
    Exit;
  end;

  LocalSize := 0;

  if fileexists(lfn) then //filesize
  begin
    LocalSize := GetFileSize(lfn);
    If LocalSize = File_Size then
    begin
      sendmsg('file exists (same size) - skipping');
      if f_retr = 'multi' then
        curpath := fn;
      FileStatus := 'File exists (same size)';
      UpdateStatus;
      exit;
    end;
  end;

//  logentry('> expected file size: '+inttostr(file_size));
  logentry('> opening output file');
  AssignFile(DataFile,LFN);
  If LocalSize > 0 then
  begin
    Reset(DataFile,1);
    Seek(DataFile, FileSize(DataFile));
  end
    ELSE
  Rewrite(DataFile,1);

  If f_retr = 'multi' then
    curpath := fn;

  CurState := 'Leeching...';
  FileStatus := '0/'+inttostr(file_size);
  UpdateStatus;

  BytesReceived := LocalSize;

  SetPASV(3);
  If Terminated then Exit;

  If LocalSize > 0 then // Local file size is greater than 0 -- Resume transfer
  begin
    SendMsg('Resuming at offset ' + inttostr(LocalSize));

    DoCommand('REST '+inttostr(LocalSize));

    If WaitForCode('350 ',30)=1 then
      DoCommand('RETR ' + FN);
  end
    ELSE
  DoCommand('RETR ' + FN);

  StartCnt := GetTickCount;

  ResetTimeout(30);

  repeat
    AmtRead := 0;
    If DataSocket.CanRead(30) then
      AmtRead := DataSocket.RecvBuffer(@DataBuf,16384)
        ELSE
      TimedOut := TRUE;
    BytesReceived := BytesReceived + AmtRead;
    If (AmtRead > 0) then
    begin
      // Data received, reset timeout timer
      ResetTimeout(30);

      blockwrite(DataFile,DataBuf,AmtRead);
      TimeCnt := GetTickCount - StartCnt;

      timetaken := timecnt / 1000;

      if timetaken > 0 then
        kbs := (bytesreceived div 1024) / timetaken
          ELSE
        kbs := 0;

      FileStatus := inttostr(bytesreceived+localsize)+'/'+inttostr(file_size)+' ('+format('%f',[(bytesreceived+localsize)/file_size*100])+'%) '+format('%f KB/sec',[kbs]);
      UpdateSpeed(kbs);
      UpdateStatus;
    end;
    CheckTimeout;
  until (DataSocket.LastError <> 0) OR Terminated OR TimedOut;

  If TimedOut then Logentry('! Timed out retrieving: '+fn);

  TimeCnt := GetTickCount - StartCnt;

  timetaken := timecnt / 1000;

  if timetaken > 0 then
    kbs := (bytesreceived div 1024) / timetaken
      ELSE         // rtlrandom sysutils
    kbs := 0;

//  sendmsg('transfer complete: '+inttostr(bytesreceived)+' bytes in '+format('%f',[timetaken])+' seconds, '+format('%f',[kbs])+'KB/sec');

  logentry('> RETR terminated: '+GetErrorDesc(DataSocket.LastError));
  logentry('> RETR received: '+inttostr(bytesreceived)+' bytes in '+format('%f',[timetaken])+' seconds, '+format('%f',[kbs])+'KB/sec');

//  If File_Size <> BytesReceived then
//    SendMsg('Filesize mismatch on '+LFN+' size='+inttostr(file_size)+' recv='+inttostr(bytesreceived));

  CurState := 'Scanning...';

  If TimedOut then
    FileStatus := 'Timed out.'
      ELSE
  If Terminated then
    FileStatus := 'Aborted.'
      ELSE
    FileStatus := 'Done.';

  UpdateStatus;
  UpdateSpeed(9999);

  CloseFile(DataFile);
end;

function FTPScanThread.ChangeDirectory(Path : string) : boolean;
begin
  Result := FALSE;
  If TimedOut OR Terminated then Exit;
  DoCommand('CWD ' + Path);
  Result := WaitForCode('250 |550 ',30) = 1;
end;

function FTPScanThread.GetDirectory(Strs : TStringList) : boolean;
var databuf : array[1..65336] of char;
    IDX, AmtRead : integer;
begin
  Result := FALSE;
  If TimedOut OR Terminated then Exit;

  MemBuffer.Clear;
  MemBuffer.Add('');

  BytesReceived := 0;

  If SetPASV(3) then
    DoCommand('LIST');

  If Terminated then
    Exit;

  WaitForCode('125 |150 ',30);

  CurState := 'Receiving...';
  UpdateStatus;

  // Process/parse data

  ResetTimeout(30);

  repeat
    AmtRead := DataSocket.RecvBuffer(@DataBuf,16384);

    BytesReceived := BytesReceived + AmtRead;
  //  logentry('br:'+inttostr(amtread));

    If AmtRead > 0 then
    begin
      ResetTimeout(30);
      for IDX := 1 to AmtRead do
        if DataBuf[IDX] in [#13,#10] then
        begin
          If DataBuf[IDX] = #13 then
            MemBuffer.Add('');
        end
          ELSE
        MemBuffer[MemBuffer.Count - 1] := MemBuffer[MemBuffer.Count - 1] + DataBuf[IDX];
    end;

    For IDX := 1 to AmtRead do
      Write(Logfile,DataBuf[IDX]);

    CheckTimeout;
//    sleep(0);
  until (DataSocket.LastError <> 0) OR Terminated OR TimedOut OR (AmtRead = -1);
  If TimedOut then LogEntry('! Timeout during LIST operation');
  LogEntry('! LIST Complete, '+inttostr(BytesReceived) + ' bytes received.');

  CurState := 'Scanning...';
  UpdateStatus;

  GetDirectory := WaitForCode('226 ',30)=1;

  Strs.Assign(MemBuffer);
end;

procedure FTPScanThread.DeleteFiles(FileSpec : string);
var TSR : TSearchRec;
begin
  If FindFirst(FileSpec,faAnyfile,TSR) = 0 then
  begin
    DeleteFile(ExtractFilePath(FileSpec)+TSR.Name);
    while findnext(tsr) = 0 do
    begin
      DeleteFile(ExtractFilePath(FileSpec)+TSR.Name);
    end;
  end;
  FindClose(TSR);
end;

procedure FTPScanThread.UpdateMain;
begin
//  form1.Memo1.lines.add(UpdateMsg); windows messages
  SendMessage(f_OwnerHWND,WM_MSGUPDATE,0,LPARAM(PChar(UpdateMsg)));
end;

procedure FTPScanThread.SendMsg(Msg : string);
begin
  If (f_OwnerHWND = 0) then Exit;

  UpdateMsg := f_SID+':'+Msg;
  UpdateMain;
end;

procedure FTPScanThread.UpdateStatus;
begin
  If (f_OwnerHWND = 0) then Exit;

  timeoutstr := format('%x',[ThreadID]);

  f_ur.si_0 := curpath;
  f_ur.si_1 := curstate;
  f_ur.si_2 := filestatus;
  f_ur.si_3 := timeoutstr;
  f_ur.tl := MainListItem;

  sendmessage(f_OwnerHWND,WM_STATUSUPDATE,0,LPARAM(@f_ur));
end;

procedure FTPScanThread.UpdateSpeed(Speed:real);
begin
  If (f_OwnerHWND = 0) then Exit;

  f_su.speed := speed;
  f_su.threadid := ThreadID;
  sendmessage(f_OwnerHWND,WM_SPEEDUPDATE,0,LPARAM(@f_su));
end;

procedure FTPScanThread.ScanDirectory(Path : string);
var DirList : TStringList;
    IDX : integer;
    TempSt : string;
begin
  // Scan a directory
  If TimedOut OR Terminated then Exit;

  If ChangeDirectory(Path) then
  begin
    CurPath := Path;
    UpdateStatus;
    Inc(RecurseCnt);
    DirList := TStringList.Create;
    DirList.Add('');
    GetDirectory(DirList);
    For IDX := 0 to DirList.Count - 1 do
      If Trim(DirList[IDX]) <> '' then
      begin
        If (DirList[IDX][1] = 'd') AND
           (trim(parsefilename(DirList[IDX])) <> '.') AND
           (trim(parsefilename(DirList[IDX])) <> '..')then // it's a directory
        begin
          ScanDirectory(Path+parsefilename(dirlist[idx])+'/');
        end
          ELSE // it's something else
        If (DirList[IDX][1] = 'l') AND
           (trim(parsefilename(DirList[IDX])) <> '.') AND
           (trim(parsefilename(DirList[IDX])) <> '..')then // it's a symbolic link
        begin
          TempSt := parsefilename(dirlist[idx]);
          TempSt := Trim(Copy(TempSt,1,pos('->',TempSt)-1));
          ScanDirectory(Path+TempSt+'/');
        end
          ELSE // it's not a directory
        begin
//          DumpToDB(parsefilename(dirlist[idx]),path,(parsefilesize(dirlist[idx]) DIV 1024));
        end;
      end;
    DirList.Free;
    Dec(RecurseCnt);
  end
    ELSE
  begin
    If Path = '/' then
    begin
      logentry('Could not CWD to root directory.');
      f_ScanStatus := 'Could not CWD root.';
      sendmsg('Thread ['+f_sid+'] could not CWD to root.');
    end
      ELSE
    logentry('Permission denied on CWD to '+Path);
  end;
end;

function FTPScanThread.CheckSocketError : boolean;
begin
  Result := FALSE;
  If DataSocket.LastError <> 0 then
  begin
    Result := TRUE;
  end;
  If ControlSocket.LastError <> 0 then
  begin
    Result := TRUE;
  end;
end;

function FTPScanThread.Login : boolean;
var Response : integer;
begin
  Result := FALSE;

  try
    ControlSocket.Connect(f_host,inttostr(f_port));
  except
    on e:exception do logentry('exception:'+e.message);
  end;

  If CheckSocketError then
  begin
    f_ScanStatus := 'Could not connect. ('+GetErrorDesc(ControlSocket.LastError)+')';
    TimedOut := pos('timed',GetErrorDesc(ControlSocket.LastError))>0;
    if pos('refused',GetErrorDesc(ControlSocket.LastError))>0 then f_ScanStatus := 'Down';
    logentry(f_ScanStatus);
    sendmsg(f_ScanStatus);
    Terminate;
    Exit;
  end;

  Response := WaitForCode('220 |530 ',30);

  If (Response = 1) AND (NOT TimedOut) then
  begin
    DoCommand('USER ' + f_User);
    Response := WaitForCode('331 |530 ',30);
    If (Response = 1) AND (NOT TimedOut) then
    begin
      DoCommand('PASS ' + f_Pass);
      Response := WaitForCode('230 |530 ',30);
      Result := (Response = 1) AND (NOT TimedOut);
    end
  end
    ELSE
  If (Response = 2) AND (NOT TimedOut) then
  begin
    f_ScanStatus := 'Denied';
  end;

  If Result then
  begin
    if f_retr='' then
    begin
      DeleteFiles('nfo\'+f_SID+'*.nfo');
      DoCommand('TYPE A');
      CurState := 'Scanning...';
      UpdateStatus;
    end
      ELSE
    DoCommand('TYPE I');
    WaitForCode('200 ',30);
  end;
end;

procedure FTPScanThread.Execute;
var idx : integer;
    tempb : boolean;
begin
  { Place thread code here }
  try

    // Open log file

    AssignFile(LogFile, extractfilepath(paramstr(0))+'logs\'+f_SID+'-thread.txt');
    Rewrite(LogFile);

    sendmsg(format('* Thread %x starting.',[ThreadID]));
    logentry(format('* Thread %x starting.',[ThreadID]));

    total_files := 0;
    total_kbytes := 0;

    CurState := 'Connecting...';
    UpdateStatus;

    CurPath := '';
    f_ScanStatus := '';

    CurState := 'Login';
    UpdateStatus;

    If Check_Status then // We're just checking the site's status to see if it's up, if we can log in/etc.
    begin
      TempB := Login;
      If TempB then f_ScanStatus := 'Up'
        ELSE
      If (NOT TempB) AND (NOT TimedOut) AND (f_ScanStatus = '') then f_ScanStatus := '???'
        ELSE
      If (NOT TempB) AND (TimedOut) then f_ScanStatus := 'Timeout';
    end
      ELSE
    If Login then // If logged in then do this stuff...  :)
    begin
      RecurseCnt := 0;

      if f_retr = '' then ScanDirectory('/')
        ELSE
      if f_retr = 'multi' then // We've been asked to retrieve multiple files
      begin
        for idx := 0 to f_flist.Count - 1 do
        begin
          RetrieveFile(extractwordl(2,f_flist[idx],'*'),extractwordl(3,f_flist[idx],'*'));
        end;
      end
        ELSE
      RetrieveFile(f_retr,f_lretr); // Retrieve a single file

      DoCommand('QUIT');

      WaitForCode('221 ',30);
    end
      ELSE
    begin
      logentry('Unable to log in.');
      if (f_ScanStatus = '') then
        f_ScanStatus := 'Unable to log in.';
      sendmsg('Thread ['+f_sid+'] was unable to log in.')
    end;

    If TimedOut then // We've timed out?  Do some logging stuff.
    begin
      logentry('Time out condition encountered');
      f_ScanStatus := 'Timed out.';
      sendmsg('Thread ['+f_sid+'] timed out')
    end
      ELSE
    begin
      if (f_ScanStatus = '') then
        f_ScanStatus := 'Completed OK.';
      sendmsg('Thread ['+f_sid+'] completed');
    end;

    logentry(format('* Thread %x terminating.',[ThreadID]));
    sendmsg(format('* Thread %x terminating.',[ThreadID]));

    // Close the log file

    CloseFile(LogFile);
  except
    on e:exception do
    begin
      sendmsg('exception:'+e.message);
      logentry('exception:'+e.message);
      f_ScanStatus := 'Exception: '+e.message;
    end;
  end;

end;

end.

